import { PublicKey, } from '@solana/web3.js';
import { getSignerHistogram, TransactionBuilder } from "../../shared";
import { RpcDriver, } from './RpcDriver';
import { FailedToConfirmTransactionError, FailedToConfirmTransactionWithResponseError, FailedToSendTransactionError, } from "../../errors";
import { zipMap } from "../../utils";
export class Web3RpcDriver extends RpcDriver {
    async sendTransaction(transaction, signers = [], sendOptions = {}) {
        var _a, _b;
        if (transaction instanceof TransactionBuilder) {
            signers = [...transaction.getSigners(), ...signers];
            transaction = transaction.toTransaction();
        }
        (_a = transaction.feePayer) !== null && _a !== void 0 ? _a : (transaction.feePayer = this.getDefaultFeePayer());
        (_b = transaction.recentBlockhash) !== null && _b !== void 0 ? _b : (transaction.recentBlockhash = await this.getLatestBlockhash());
        signers = [this.metaplex.identity(), ...signers];
        const { keypairs, identities } = getSignerHistogram(signers);
        if (keypairs.length > 0) {
            transaction.partialSign(...keypairs);
        }
        for (let i = 0; i < identities.length; i++) {
            await identities[i].signTransaction(transaction);
        }
        const rawTransaction = transaction.serialize();
        try {
            return await this.metaplex.connection.sendRawTransaction(rawTransaction, sendOptions);
        }
        catch (error) {
            // TODO: Parse using program knowledge when possible.
            throw new FailedToSendTransactionError(error);
        }
    }
    async confirmTransaction(signature, commitment) {
        let rpcResponse;
        try {
            rpcResponse = await this.metaplex.connection.confirmTransaction(signature, commitment);
        }
        catch (error) {
            throw new FailedToConfirmTransactionError(error);
        }
        if (rpcResponse.value.err) {
            throw new FailedToConfirmTransactionWithResponseError(rpcResponse);
        }
        return rpcResponse;
    }
    async sendAndConfirmTransaction(transaction, signers, confirmOptions) {
        const signature = await this.sendTransaction(transaction, signers, confirmOptions);
        const confirmResponse = await this.confirmTransaction(signature, confirmOptions === null || confirmOptions === void 0 ? void 0 : confirmOptions.commitment);
        return { signature, confirmResponse };
    }
    async getAccount(publicKey, commitment) {
        const accountInfo = await this.metaplex.connection.getAccountInfo(publicKey, commitment);
        return this.getUnparsedMaybeAccount(publicKey, accountInfo);
    }
    async getMultipleAccounts(publicKeys, commitment) {
        const accountInfos = await this.metaplex.connection.getMultipleAccountsInfo(publicKeys, commitment);
        return zipMap(publicKeys, accountInfos, (publicKey, accountInfo) => {
            return this.getUnparsedMaybeAccount(publicKey, accountInfo);
        });
    }
    async getLatestBlockhash() {
        return (await this.metaplex.connection.getLatestBlockhash('finalized')).blockhash;
    }
    getDefaultFeePayer() {
        const identity = this.metaplex.identity().publicKey;
        return identity.equals(PublicKey.default) ? undefined : identity;
    }
    getUnparsedMaybeAccount(publicKey, accountInfo) {
        if (!accountInfo) {
            return { publicKey, exists: false };
        }
        return { publicKey, exists: true, ...accountInfo };
    }
}
//# sourceMappingURL=Web3RpcDriver.js.map